/* Copyright (C) 2006-2008 Peter Selinger.  
   This file is part of unixcrypt-breaker. It is free software and it
   is distributed under the terms of the GNU general public license.
   There is no warranty. See the file COPYING for details. */

/* ---------------------------------------------------------------------- */
/* user interface */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <getopt.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "stats.h"
#include "block.h"

#define NAME "unixcrypt-breaker"

/* ---------------------------------------------------------------------- */
/* print usage information */

static void usage(FILE *fout) {
  fprintf(fout, ""NAME" "VERSION". Break unix \"crypt\" encryption.\n\n");
  fprintf(fout, 
  "Usage: "NAME" [options] -s <statfile> -c <corpusfile>...\n"
  "       "NAME" [options] -s <statfile> <cipherfile>\n"
  "\n"
  "The code-breaking algorithm relies on statistical information from a\n"
  "corpus of sample data that is presumed similar to the unknown\n"
  "plaintext. To use this program, you must first run "NAME"\n"
  "with the -c option on a corpus of sample data to create a statfile\n"
  "(see the first form of usage above). You may then use the resulting\n"
  "statfile to attempt decryption of the cipherfile (second form).\n"
  "\n"
  "Options:\n"
  " -h, --help            print this help message and exit\n"
  " -V, --version         print version info and exit\n"
  " -L, --license         print license info and exit\n"
  " -D, --debug           raise debugging level\n"
  " -c, --corpus          create statfile from one or more files of sample\n"
  "                       data, or from stdin if no filenames are given.\n"
  " -s, --statfile <file> specify path to statfile (mandatory)\n"
  " -d, --display <mode>  progress display: color (default), mono, scroll, or none\n"
  " -o, --output <file>   select output file (default: <cipherfile>.guess)\n"
  " -w, --wirefile <file> output wiring information (advanced) to file\n"
  " --                    end of options, filenames follow\n");
}

/* print version and copyright information */
static void version(FILE *fout) {
  fprintf(fout, ""NAME" "VERSION". Break unix \"crypt\" encryption.\n");
  fprintf(fout, "Copyright (C) 2006-2008 Peter Selinger.\n");
}

static void license(FILE *fout) {
  fprintf(fout, ""NAME" "VERSION". Break unix \"crypt\" encryption.\n");
  fprintf(fout, "Copyright (C) 2006-2008 Peter Selinger.\n");
  fprintf(fout, "\n");
  fprintf(fout,
  "This program is free software; you can redistribute it and/or modify\n"
  "it under the terms of the GNU General Public License as published by\n"
  "the Free Software Foundation; either version 2 of the License, or\n"
  "(at your option) any later version.\n"
  "\n"
  "This program is distributed in the hope that it will be useful,\n"
  "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
  "GNU General Public License for more details.\n"
  "\n"
  "You should have received a copy of the GNU General Public License\n"
  "along with this program; if not, see <http://www.gnu.org/licenses/>.\n"
  );
}

static struct option longopts[] = {
  {"help",         0, 0, 'h'},
  {"version",      0, 0, 'V'},
  {"license",      0, 0, 'L'},
  {"corpus",       0, 0, 'c'},
  {"statfile",     1, 0, 's'},
  {"output",       1, 0, 'o'},
  {"display",      1, 0, 'd'},
  {"debug",        0, 0, 'D'},
  {"wirefile",     1, 0, 'w'},
  {0, 0, 0, 0}
};

static char *shortopts = "hVLcs:o:d:Dw:";

static int digest_corpus(char *statfile, char **av, int ac) {
  FILE *fstat;
  FILE *fin;
  int i;
  stats_t *t;
  
  /* check for statfile */
  if (statfile == NULL) {
    fprintf(stderr, ""NAME": the -s option is mandatory. Try --help for more info.\n");
    exit(1);
  }
  fstat = fopen(statfile, "w");
  if (!fstat) {
    fprintf(stderr, ""NAME": %s: %s\n", statfile, strerror(errno));
    exit(1);
  }      

  t = stats_new();
  if (!t) {
    fprintf(stderr, ""NAME": %s\n", strerror(errno));
    exit(1);
  }
  
  if (ac == 0) {  /* read from stdin */
    stats_digest(stdin, t);
  } else {
    for (i=0; i<ac; i++) {
      fin = fopen(av[i], "r");
      if (!fin) {
	fprintf(stderr, ""NAME": %s: %s, skipping\n", av[i], strerror(errno));
      } else {
	stats_digest(fin, t);
	fclose(fin);
      }
    }
  }
  stats_write(fstat, t);
  stats_free(t);
  fclose(fstat);
  return 0;
}

/* wirefile is optional */
static int unixcrypt_breaker(char *cipherfile, char *statfile, char *wirefile, char *outfile, int display_mode, int debug) {

  FILE *fcipher;
  FILE *fwire;
  FILE *fstat;
  FILE *fout;
  int n, b;
  int block_cipher[256];
  int block_plain[256];
  int wires_known[256];
  stats_t *fr;

  /* open and read statfile */
  if (statfile == NULL) {
    fprintf(stderr, ""NAME": the -s option is mandatory. Try --help for more info.\n");
    exit(1);
  }
  fstat = fopen(statfile, "r");
  if (!fstat) {
    fprintf(stderr, ""NAME": %s: %s\n", statfile, strerror(errno));
    exit(1);
  }      

  fr = stats_read(fstat);
  if (!fr) {
    fprintf(stderr, ""NAME": %s\n", strerror(errno));
    return 1;
  } else if (fr == STATS_EFORMAT) {
    fprintf(stderr, ""NAME": %s: bad file format\n", statfile);
    return 1;
  }
  fclose(fstat);

  /* open cipherfile */
  fcipher = fopen(cipherfile, "r");
  if (!fcipher) {
    fprintf(stderr, ""NAME": %s: %s\n", cipherfile, strerror(errno));
    exit(1);
  }

  /* open outfile */
  fout = fopen(outfile, "w");
  if (!fout) {
    fprintf(stderr, ""NAME": %s: %s\n", outfile, strerror(errno));
    exit(1);
  }

  /* open wirefile, if any */
  if (wirefile) {
    fwire = fopen(wirefile, "w");
    if (!fwire) {
      fprintf(stderr, ""NAME": %s: %s\n", wirefile, strerror(errno));
      exit(1);
    }
  } else {
    fwire = NULL;
  }

  b=0;
  while (1) {
    n = read_block(fcipher, block_cipher);
    if (n==0) {
      break;
    }
    unmask_block(n, block_cipher);
    erase_block(block_plain);
    process_block(fr, n, block_cipher, block_plain, wires_known, display_mode, debug);
    if (display_mode == DISPLAY_SCROLL) {
      write_block_hex(stdout, n, block_plain, NULL, 0);
      fprintf(stdout, "\n");
    }
    write_block(fout, n, block_plain);
    fflush(fout);
    if (fwire) {
      write_wires_long(fwire, wires_known, b);
      fflush(fwire);
    }
    b++;
    if (n<256) {
      break;
    }
  }

  fclose(fcipher);
  fclose(fout);
  if (fwire) {
    fclose(fwire);
  }
  return 0;
}

int main(int ac, char *av[]) {
  int c;
  char *cipherfile;
  const char *guessext = ".guess";

  /* command line options and defaults */
  int display_mode = DISPLAY_COLOR;  
  int corpus_mode = 0;   /* 1=on, 0=off */
  char *statfile = NULL;
  char *wirefile = NULL;
  char *outfile = NULL;
  int debug = 0;

  while ((c = getopt_long(ac, av, shortopts, longopts, NULL)) != -1) {
    switch (c) {
    case 'h':
      usage(stdout);
      exit(0);
      break;
    case 'V':
      version(stdout);
      exit(0);
      break;
    case 'L':
      license(stdout);
      exit(0);
      break;
    case 'c':
      corpus_mode = 1;
      break;
    case 's':
      statfile = optarg;
      break;
    case 'o':
      outfile = optarg;
      break;
    case 'd':
      if (strcmp(optarg, "color") == 0) {
	display_mode = DISPLAY_COLOR;
      } else if (strcmp(optarg, "mono") == 0) {
	display_mode = DISPLAY_MONO;
      } else if (strcmp(optarg, "scroll") == 0) {
	display_mode = DISPLAY_SCROLL;
      } else if (strcmp(optarg, "none") == 0) {
	display_mode = DISPLAY_NONE;
      } else {
	fprintf(stderr, ""NAME": invalid display mode: %s. Try --help for more info.\n", optarg);
	exit(1);
      }
      break;
    case 'D':
      debug++;
      break;
    case 'w':
      wirefile = optarg;
      break;
    case '?': default:
      fprintf(stderr, "Try --help for more information.\n");
      exit(1);
      break;
    }
  }

  if (corpus_mode) {
    return digest_corpus(statfile, &av[optind], ac - optind);
  }

  if (ac - optind != 1) {
    fprintf(stderr, ""NAME": no filenames specified. Try --help for more info.\n");
    exit(1);
  }

  cipherfile = av[optind];

  if (outfile == NULL) {
    outfile = malloc(strlen(cipherfile)+strlen(guessext)+1);
    if (!outfile) {
      fprintf(stderr, ""NAME": %s\n", strerror(errno));
      exit(1);
    }
    strcpy(outfile, cipherfile);
    strcat(outfile, guessext);
  }

  return unixcrypt_breaker(cipherfile, statfile, wirefile, outfile, display_mode, debug);
}

